package agents.TheNegotiator;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Random;
import negotiator.Bid;

/**
 * The BidsCollection class stores the bids of the partner and all possible bids.
 * 
 * @author Alex Dirkzwager, Mark Hendrikx, Julian de Ruiter
 */
public class BidsCollection {

	// bids done by partner
	private ArrayList<UTBid> partnerBids;
	// all possible bids (for us) which do not violate the constraints
	private ArrayList<UTBid> possibleBids;

	/**
	 * Creates a BidsCollection-object which stores the partner bids and all possible
	 * bids.
	 */
	public BidsCollection() {
		partnerBids = new ArrayList<UTBid>();
		possibleBids  = new ArrayList<UTBid>();
	}

	/**
	 * @return the partnerBids
	 */
	public ArrayList<UTBid> getPartnerBids() {
		return partnerBids;
	}

	/**
	 * @return the possibleBids
	 */
	public ArrayList<UTBid> getPossibleBids() {
		return possibleBids;
	}

	/**
	 * Add a partner bid to the history. Bids are stored at the front
	 * to preserve the timeline.
	 * 
	 * @param bid made by partner in the current turn
	 * @param utility of the bid
	 */
	public void addPartnerBid(Bid bid, double utility) {
		UTBid utbid = new UTBid(bid, utility);
		partnerBids.add(0, utbid);
	}
	
	/**
	 * Add a possible bid to the list of possible bids. The given bid 
	 * should not violate the constraints of the negotiation.
	 * 
	 * @param bid which is possible
	 * @param utility of the bid
	 */
	public void addPossibleBid(Bid bid, double utility) {
		UTBid utbid = new UTBid(bid, utility);
		possibleBids.add(utbid);
	}
	
	/**
	 * Sorts all possible bids in reverse natural order.
	 */
	public void sortPossibleBids() {
		Collections.sort(possibleBids);
	}
		
	/**
	 * Get a partner bid.
	 * 
	 * @param i 
	 * @return the i'th bid in the timeline
	 */
	public Bid getPartnerBid(int i) {
		Bid bid = null;
		
		if (i < partnerBids.size()) {
			bid = partnerBids.get(i).getBid();
		} else {
			ErrorLogger.log("BIDSCOLLECTION: Out of bounds");
		}
		return bid;
	}
		
	/**
	 * Return the mean of the X bids before this bid.
	 * 
	 * @param the last X bids which should be considered
	 * @return sum of previous bids
	 */
	public double meanPartnerBids(int lastX) {
		double sum = 0;
		// skip the current observation
		for (int i = 1; i <= lastX; i++) {
			sum += partnerBids.get(i).getUtility();
		}
		return (double)(sum / lastX);
	}
	
	/**
	 * Return the variance of the X bids before this bid.
	 * 
	 * @param the last X bids which should be considered
	 * @return variance of previous X bids
	 */
	public double variancePartnerBids(int lastX) {
		// unbiased variance estimator => sum((x_i - mean)^2) / (n - 1)
		double variance = 0;
		if (lastX > 1) {
			double mean = meanPartnerBids(lastX);
			double sumXiMinusMean = 0;
			for (int i = 0; i <lastX; i++) {
				sumXiMinusMean += Math.pow((partnerBids.get(i).getUtility() - mean), 2);
			}
			variance = sumXiMinusMean / ((double)(lastX - 1));
		}
		return variance;
	}

	/**
	 * Get a partner bid which has a utility of at least a certain
	 * value. Null is returned if no such bid exists.
	 * 
	 * @param threshold
	 * @return bid with utility > threshold if exists
	 */
	public Bid getBestPartnerBids(double threshold) {
		ArrayList<UTBid> temp = partnerBids;
		Collections.sort(temp);
		Bid bid;

		int count = 0;
		while (count < temp.size() && temp.get(count).getUtility() >= threshold) {
			count++;
		}
		
		if (count > 0) {
			Random randomnr = new Random();
			bid = temp.get(randomnr.nextInt(count)).getBid();
		} else {
			bid = null;
		}
		return bid;
	}

	/**
	 * Get a random bid between two given thresholds.
	 * 
	 * @param lowerThres lowerbound threshold
	 * @param upperThres upperbound threshold
	 * @return random bid between thresholds
	 */
	public Bid getOwnBidBetween(double lowerThres, double upperThres) {
		int lB = 0;
		int uB = 0;
		Bid bid;

		// determine upperbound and lowerbound by visiting all points
		for (int i = 0; i < possibleBids.size(); i++) {
			double util = possibleBids.get(i).getUtility();
			if (util >= upperThres) {
				uB++;
			}
			if (util >= lowerThres) {
				lB++;
			}
		}
		// if there are no points between the bounds
		if (lB == uB) {
			// ignore upper threshold
			bid = getOwnBidBetween(lowerThres, 1.1);
		} else {
			// decrement upper- and lowerbound to get the correct index
			// (count counts from 1, while arrays are indexed from 0)
			if (lB > 0) {
				lB--;
			}
			if ((uB + 1) <= lB) {
				uB++;
			}
			// calculate a random bid index
			int result = uB + (int) ( Math.random() * (lB - uB) + 0.5);
			bid = possibleBids.get(result).getBid();
		}
		return bid;
	}
}