package negotiation.AgentX;

import java.util.ArrayList;

import negotiator.Bid;
import negotiator.Domain;
import negotiator.issue.IssueDiscrete;
import negotiator.issue.Value;
import negotiator.issue.ValueDiscrete;
import negotiator.utility.UtilitySpace;

/**
 * Class for building an opponent model in discrete space.
 * Contains value- and issue-processors to deal with opponent bids. 
 * 
 * @author E. Jacobs
 *
 */
public class OpponentModel {

	//Set utility space, with issues
	private UtilitySpace util;
	private Domain dom;
	private DiscreteIssueProcessor issueProcessor;
	
	private ArrayList<IssueDiscrete> discreteIssueList = new ArrayList<IssueDiscrete> ();
	private ArrayList<DiscreteValueProcessor> valueProcessList = new ArrayList<DiscreteValueProcessor> ();;
	
	/**
	 * Creates an opponent model for the given utility space
	 * @param u The utility space
	 */
	public OpponentModel(UtilitySpace u){
		//System.out.println("Building opponent model...");
		util = u;
		dom = util.getDomain();
		issueProcessor = new DiscreteIssueProcessor(dom);
		discreteIssueList = issueProcessor.getIssueList();
		createValueProcessors();
		//System.out.println("Opponent model built");
	} 
	
	/**
	 * Creates discreteValueProcessors for all issues
	 */
	private void createValueProcessors(){
		
		for (IssueDiscrete i : discreteIssueList){
			valueProcessList.add(new DiscreteValueProcessor(i.getValues()));
		}
	}
	
	/**
	 * Returns a DiscreteValueProcessor, allowing the use of several methods 
	 * @param i The issue for which a ValueProcessor is needed
	 * @return a DiscreteValueProcessor for the issue
	 */
	public DiscreteValueProcessor getValueProcessor(IssueDiscrete i){
		
		int index = discreteIssueList.indexOf(i);
		
		if (index != -1) return valueProcessList.get(index);
		
		return null;
		
	}
	
	/**
	 * Gives the discreteIssueProcessor for this opponent model
	 * @return The discreteIssueProcessor
	 */
	public DiscreteIssueProcessor getIssueProcessor(){
		return issueProcessor;
	}
	
	/**
	 * Processes a bid, possibly changing value ranks for the internal opponent model. Currently, values on which more bids are made are ranked higher.
	 * @param b The bid done by the opponent
	 * @param time Time at which the bid was done
	 */
	public void processBid(Bid b, double time){
		
		issueProcessor.adaptWeightsByBid(b, time);
		//System.out.println(issueProcessor.toString());
			
		for (IssueDiscrete i : discreteIssueList){
			
			//System.out.println("Dealing with issue:" + i.getName());
			
			Value v = null;	
			
			try {
				v = b.getValue(i.getNumber());
			} catch (Exception e) {}
			
			if (v != null){
				ValueDiscrete vDisc = (ValueDiscrete) v;
				getValueProcessor(i).addBidForValue(vDisc);
				//System.out.println(getValueProcessor(i).toString());
			}			
		}
	}

	/**
	 * Gives the normalized value rank of some value within a certain Issue
	 * @param issueIndex The index of the issue. Same index as in the ArrayList provided by 
	 * utilitySpace.getDomain().getIssues()
	 * @param valueOfIssue The value from that issue for which the normalized rank is required
	 * @return The normalized valueRank for the value of the issue given
	 */
	public double getEvaluationOfValue(int issueIndex, ValueDiscrete valueOfIssue) {
			return valueProcessList.get(issueIndex).getNormalizedValueRank(valueOfIssue);
	}
}
