package negotiation.group4;

import java.io.IOException;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Stack;

import negotiator.Agent;
import negotiator.Bid;
import negotiator.BidIterator;
import negotiator.actions.Accept;
import negotiator.actions.Action;
import negotiator.actions.Offer;

import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.*;

/**
 * Negotiation agent developed by Group 4 for IN4010 Final Assignment.
 * 
 * The agent is capable of selecting a negotiation strategy using a {@link}StrategyAnalyzer.
 * Constructs an opponent model that can be used by strategies to construct optimal bids.
 * 
 * @author Lennaert van den Brink <l.o.vandenbrink@student.tudelft.nl>
 * @author Hans Janmaat <c.j.janmaat@student.tudelft.nl>
 * @author Cees-Willem Hofstede <c.w.p.l.j.hofstede@student.tudelft.nl>
 * @author Berend Klein Haneveld <b.a.kleinhaneveld@student.tudelft.nl>
 */
public class Group4 extends Agent {
	
	// generator for unique id (useful in logging)
	private static final AtomicLong UNIDGENERATOR = new AtomicLong();
	
	/*
	 * Use this variable to enable/disable logging in total
	 * This does not set the logging level if enabled.
	 */
	private static boolean enableLogging = false;
	
	// private logger which can be called statically or trough agent functions (to identify agent)
	public static Logger logger = Logger.getLogger("negotiation.group4.group4");
	static {
		if(enableLogging)
		{
			logger.setLevel(Level.FINEST); // SEVERE, WARNING, INFO, CONFIG, FINE, FINER or FINEST 
		    try {
		    	boolean append = true;
		    	FileHandler fh = new FileHandler("log/group4.xml", append);
		    	fh.setFormatter(new XMLFormatter());
		    	logger.addHandler(fh);
		    }
		    catch (IOException e) {
		    	e.printStackTrace();
		    }
		}
		else
		{
			// do not set up a handler for the logger to print to a file
			// also disable the root logging (ie to console)
			logger.setUseParentHandlers(false);
		}
	}
	
	// Opponent model
	public OpponentModel opponentModel;
	
	private Strategy opponentStrategy;	
	
	// Data structures for memorized actions
	public Stack<ExtendedBid> thisAgentBids;			// All bids done by Group4 agent
	public ArrayList<ExtendedBid> rankedBids = null;	// All possible bids ranked by utility
	public int lastBidIndex;							// index in renkedBids of agents last bid
	
	// unique id
	private String unid;
	
	/**
	 * Constructor
	 */
	public Group4() {
		unid = "Agent " + Long.toString(UNIDGENERATOR.getAndIncrement()+1);
		thisAgentBids = new Stack<ExtendedBid>();
		opponentStrategy = new Strategy(this);
		
		debug("New Group4 session started");
	}
	
	/**
	 * This function is called when a next session starts with the same opponent.
	 */
	public void init() 
	{
		// Re-initialize all arrays, objects and values
		AssertOpponentModel();
		
		opponentModel.init();
		rankAllBids();
	}
	
	/**
	 * Make sure the opponent model is set
	 * 
	 */
	public void AssertOpponentModel() {
		if (opponentModel == null) {
			opponentModel = new OpponentModel(this);
		}
	}
	
	/**
	 * @return Name of this agent
	 */
	@Override
	public String getName()
	{
		return "Dread Pirate Roberts";
	}
	
	/**
	 * Returns version number
	 * @return String with version number
	 */
	public static String getVersion() {
		return "42";
	}
	
	/**
	 * Stores opponent action in array list and prints the 
	 * action if it differs from previous actions.
	 * 
	 * @param opponentAction Action from opponent
	 */
	public void ReceiveMessage(Action opponentAction) 
	{
		// If the opponent action is an offer, save it for later 
		if (opponentAction instanceof Offer) {
			Offer opponentOffer = (Offer)opponentAction;
			
			try {
				// Add the bid to the list of opponent bids
				ExtendedBid opponentBid = new ExtendedBid(opponentOffer.getBid(), utilitySpace.getUtility(opponentOffer.getBid()));
				opponentModel.bids.add(opponentBid);
				
				// Learn from the opponent if the bid is unique
				if (opponentModel.isUniqueBid(opponentBid)) {
					opponentModel.opponentOffered(opponentOffer);
					opponentModel.uniqueBids.add(opponentBid);;
				}
				
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * @return Action to perform
	 */
	@Override
	public Action chooseAction() {
		Action action = null;
		
		// Use the strategy to define new action
		try {
			action = opponentStrategy.getAction();
			// Save the offer if it is an offer
			if (action instanceof Offer) {
				thisAgentBids.push(new ExtendedBid(((Offer) action).getBid(), 0));
			}
		} catch (Exception e) {
			e.printStackTrace();
			error(e.getStackTrace().toString());
			// Ohnothecodehasfailed:accept
			action = new Accept(getAgentID());
		}
		
		return action;
	}
	
	/**
	 * Use the BidIterator to find all possible bids.
	 * Then sort the bids from highest utility to lowest
     */
	private void rankAllBids() {
		rankedBids = new ArrayList<ExtendedBid>();

		// Fill the list with bids and their utilities
		BidIterator bidIterator = new BidIterator(utilitySpace.getDomain());
		while (bidIterator.hasNext()) {
			Bid bid = bidIterator.next();
			double utility = 0.0;
			try {
				utility = utilitySpace.getUtility(bid);
			} catch (Exception e) {
				error(e.getStackTrace().toString());
			}

			rankedBids.add(new ExtendedBid(bid, utility));
		}

		// Sort the bids: highest utility first
		Collections.sort(rankedBids);
	}
	
	/*
	 * Get a Bid based on the index in rankedBids and set lastBidIndex.
	 */
	public ExtendedBid getOfferBid(int index) {
		lastBidIndex = index;
		return rankedBids.get(index);
	}
	
	/**
	 * Returns the index of the bid closest to the latest bid within given epsilon (this can be positive or negative).
	 * If no bid was found return last bid.
	 * Remember: the bids are sorted on utility from high to low!
	 * @param epsilon How much can the new utility differ from the last bid
	 * @return Index of first bid with utility within range or last of none found
	 */
	public int getIndexOfBidCloseToLastUtility(double epsilon) {
		int bidIndex = lastBidIndex;
		ExtendedBid closestBid;
		if(epsilon < 0 && lastBidIndex < rankedBids.size()-1)
		{
			closestBid = rankedBids.get(lastBidIndex+1);
			if(closestBid.utility + epsilon > thisAgentBids.lastElement().utility)
				bidIndex = lastBidIndex+1;
		}
		else if(epsilon > 0 && lastBidIndex > 0)
		{
			closestBid = rankedBids.get(lastBidIndex-1);
			if(closestBid.utility + epsilon < thisAgentBids.lastElement().utility)
				bidIndex = lastBidIndex-1;
		}
		
		// result must be at least 0
		return bidIndex;
	}
	
	
	/* Private stuff */
	
	//--------------------------------------------------------------------------------------------------------------
	
	/******************************************************\
	 * 					LOGGING FUNCTIONS				  *
	\******************************************************/
	
	/**
	 * Use logger to debug a message.
	 * Uses the Level FINE
	 * 
	 * @param msg
	 */
	public void debug(String msg) {
		log(Level.FINE, msg);
	}
	
	/**
	 * Use logger to show info.
	 * Uses the Level INFO
	 * 
	 * @param msg
	 */
	public void info(String msg) {
		log(Level.INFO, msg);
	}
	
	/**
	 * Use logger to show an error message.
	 * Uses the Level SEVERE
	 * 
	 * @param msg
	 */
	public void error(String msg) {
		log(Level.SEVERE, msg);
	}
	
	/**
	 * Use logger to log message, with the unid prepended to the message.
	 * 
	 * @param level
	 * @param msg
	 */
	public void log(Level level, String msg) {
		if(enableLogging)
		{
			logger.log(level, unid + ": " + msg);
		}
	}
}
