package negotiation.group4;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.util.Date;
import java.text.SimpleDateFormat;

import javax.xml.parsers.*;

import org.w3c.dom.*;

/**
 * This class is used to parse the logfiles created during tournaments. It goes
 * trough all the non-extensive logfiles and loops over all of the available
 * NegotiationOutcome nodes. It will create a comma separated value (csv)
 * document containing per opponent of the given agent. The name of this agent
 * can be given as a commandline parameter. Groep4 agent will be taken as default.
 * 
 * This is a main class, so it will in fact create a {@link}AgentLogParser object
 * to read and write the files.
 * 
 * @author Lennaert van den Brink <l.o.vandenbrink@student.tudelft.nl>
 * @author Hans Janmaat <c.j.janmaat@student.tudelft.nl>
 * @author Cees-Willem Hofstede <c.w.p.l.j.hofstede@student.tudelft.nl>
 * @author Berend Klein Haneveld <b.a.kleinhaneveld@student.tudelft.nl>
 * 
 */
public class LogParser {
	/**
	 * The name of the agent to analyze
	 */
	public static String agent = "Groep4";
	
	/**
	 * The filename of the csv fiole to be created
	 */
	public static String csvfile;

	/**
	 * @param args List with arguments. Currently supported are:
	 *  	  - arg[0]: the name of the agent, default is Groep4
	 *  	  - arg[1]: the path to the csv file, default is analysis/[agentname]_utility_[datetime].csv
	 */
	public static void main(String[] args) {
		if (args.length > 0) {
			agent = args[0];
		}
		if (args.length > 1) {
			csvfile = args[1];
			System.out.println(agent);
			System.out.println(csvfile);
		}
		else {
			//get current date to use in filename
			Date dNow = new Date();
		    SimpleDateFormat ft = new SimpleDateFormat ("yyyy-MM-dd_hh-mm-ss");
		    csvfile = "src/analysis/" + agent + "_utilities_" + ft.format(dNow) +  ".csv";
		}
		println("LogParser started for agent " + agent);
		AgentLogParser alp = new AgentLogParser(agent);
		alp.createCSV(csvfile);
		println("Done, file saved to " + csvfile);
	}
	
	/**
	 * shortcut to System.out.println
	 * 
	 * @param msg
	 */
	public static void println(String msg) {
		System.out.println(msg);
	}
}

/**
 * The actual parser, {@see}LogParser
 * 
 * @author Lennaert van den Brink <l.o.vandenbrink@student.tudelft.nl>
 * @author Hans Janmaat <c.j.janmaat@student.tudelft.nl>
 * @author Cees-Willem Hofstede <c.w.p.l.j.hofstede@student.tudelft.nl>
 * @author Berend Klein Haneveld <b.a.kleinhaneveld@student.tudelft.nl>
 *
 */
class AgentLogParser {
	/**
	 * The name of the agent to analyze
	 */
	private String agent;
	
	/**
	 * constructor
	 * 
	 * @param sAgent The name of the agent to analyze
	 */
	public AgentLogParser(String sAgent) {
		agent = sAgent;
	}
	
	/**
	 * Create a csv file for the agent. It will loop over de log
	 * directory and call getCSVLines for each of the valid files.
	 * A bug in Genius causes some of the files to mis a valid end
	 * tag, which renders invalid xml. Please make sure to edit the
	 * files to include a valid end tag (</a>).
	 * 
	 * @param csvfilePath
	 */
	public void createCSV(String csvfilePath) {
		// get the files in the log directory
		File[] files = new File("src/log").listFiles();
		
		// create a writeable file
		FileWriter csvfilestream;
		try {
			csvfilestream = new FileWriter(csvfilePath);
			BufferedWriter csvfile = new BufferedWriter(csvfilestream);
			
			// print the headers for each column
			String CSVHeader = getCSVField("OPPONENT NAME") + 
							   getCSVField("OPPONENT CLASS") + 
							   getCSVField("AGENT " + agent + " WON") + 
							   getCSVField("AGENT " + agent + " UTILITY") + 
							   getCSVField("OPPONENT UTILITY") + 
							   getCSVField("NR OF BIDS") +
							   getCSVField("DOMAIN") +
							   getCSVField("AGENT " + agent + " UTILITY SPACE") +
							   getCSVField("OPPONENT UTILITY SPACE") +
							   getCSVField("ERRORS", true);
			CSVHeader += "\n";
			csvfile.write(CSVHeader);
			
			// for each file, check whether it is a valid logfile
			// and if so, get csv from it.
			for (File file : files) {
				String fileName = file.getName();
				System.out.println("Parsing file: " + fileName);
				//TODO replace dirty check on length of filename
		        if (!file.isDirectory() && fileName.length() == 23) {
		        	csvfile.write(getCSVLines(file.getPath()));
		        }
		    }
			
			csvfile.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Will read the given file and extract useful information
	 * for the agent to be analyzed. 
	 * 
	 * @param fileURL The path to the file
	 * @return String The lines of csv extracted from the file
	 */
	private String getCSVLines(String fileURL) {
		// set empty line (for in case something goes wrong, still a
		// valid String can be returned
		String csvlines = "";
		try {
			InputStream filestream = new FileInputStream(fileURL);
			DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
			DocumentBuilder db = dbf.newDocumentBuilder();
			Document doc = db.parse(filestream);
			
			// get a list of all outcomes per session
			NodeList outcomes = doc.getElementsByTagName("NegotiationOutcome");

			int nrOutcomes = outcomes.getLength();
			for (int outcomeIndex = 0; outcomeIndex < nrOutcomes; outcomeIndex++) {
				// for each of the outcomes, call getCSVLine to extract information
				// and add it to the csv
				Node outcome = outcomes.item(outcomeIndex);
				String outcomeCSVLine = getCSVLine(outcome);
				csvlines += outcomeCSVLine != "" ? outcomeCSVLine + "\n" : "";
			}
			
			// close the file, we're done
			filestream.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		return csvlines;
	}
	
	/**
	 * Get the csv line from a single session.
	 * 
	 * @param outcome The XML node containing the information for the line of csv
	 * @return
	 */
	private String getCSVLine(Node outcome) {
		String CSVLine = "";
		
		// flag to see whether the analyzed agent took part in this session
		boolean containsAgent = false;
		
		// get the attributes of this session
		NamedNodeMap outcomeAttributes = outcome.getAttributes();
		
		// get the nodes with information per agent in this session
		NodeList outcomeChildNodes = outcome.getChildNodes();

		//TODO uses magic numbers now, this should check on node type/name
		NamedNodeMap agentANodeAttributes = outcomeChildNodes.item(1).getAttributes();
		NamedNodeMap agentBNodeAttributes = outcomeChildNodes.item(3).getAttributes();
		
		// "pointers" to correct NamedNodeMap
		NamedNodeMap agentAttributes = null;
		NamedNodeMap opponentAttributes = null;
		
		// set pointers to prevent duplicate code later on
		if(agentANodeAttributes.getNamedItem("agentName").getNodeValue().equals(agent)) {
			agentAttributes = agentANodeAttributes;
			opponentAttributes = agentBNodeAttributes;
			containsAgent = true;
		}
		else if(agentBNodeAttributes.getNamedItem("agentName").getNodeValue().equals(agent)) {
			agentAttributes = agentBNodeAttributes;
			opponentAttributes = agentANodeAttributes;
			containsAgent = true;
		}
		
		if (containsAgent)
		{
			// if the agent took part in this session, get all information from the
			// xml nodes and create a csv line. 
			String errors			  = "";
			String domain 			  = outcomeAttributes.getNamedItem("domain").getNodeValue();
			String nrOfBids			  = outcomeAttributes.getNamedItem("bids").getNodeValue();
			String agentUtility       = agentAttributes.getNamedItem("finalUtility").getNodeValue();
			String agentUtilSpace     = agentAttributes.getNamedItem("utilspace").getNodeValue();
			String opponentName       = opponentAttributes.getNamedItem("agentName").getNodeValue();
			String opponentClass      = opponentAttributes.getNamedItem("agentClass").getNodeValue();
			String opponentUtility    = opponentAttributes.getNamedItem("finalUtility").getNodeValue();
			String opponentUtilSpace  = opponentAttributes.getNamedItem("utilspace").getNodeValue();
			String agentWon			  = Float.valueOf(agentUtility.trim()).floatValue() > Float.valueOf(opponentUtility.trim()).floatValue() ? "1": "0";
			
			// errors is not always set, but if it is, print it in the csv
			Node errorNode = outcomeAttributes.getNamedItem("errors");
			if(errorNode != null) {
				errors = errorNode.getNodeValue();
			}

			CSVLine = getCSVField(opponentName) + 
					  getCSVField(opponentClass) + 
					  getCSVField(agentWon) +
					  getCSVField(agentUtility) + 
					  getCSVField(opponentUtility) + 
					  getCSVField(nrOfBids) + 
					  getCSVField(domain) +
					  getCSVField(agentUtilSpace) +
					  getCSVField(opponentUtilSpace) +
					  getCSVField(errors, true);
		} else {
			System.out.println("Does not contain name of agent");
		}
		
		return CSVLine;
	}
	
	private String getCSVField(String value) {
		return getCSVField(value, false);
	}
	
	/**
	 * Get a valid csv representation for a given string
	 * @param value
	 * @param last If true, do not add a comma at the end
	 * @return String CSV representation
	 */
	private String getCSVField(String value, boolean last) {
		String field = "\"" + value + "\"";
		if(!last) {
			field += ",";
		}
		return field;
	}
}
