package negotiation.group4;

import java.util.ArrayList;

import negotiator.Bid;
import negotiator.actions.Offer;
import negotiator.issue.Issue;
import negotiator.issue.IssueDiscrete;
import negotiator.issue.ValueDiscrete;

/**
 * Class to hold the model for the opponent, saved in a matrix from which
 * the preferences of issues can be extracted. Provides handles to get
 * bids that are slightly lower/higher than the given bid iot concede or
 * retract. Useful for strategies to define their next bid.
 * 
 * The matrix is called the tracker: bid issue tracker matrix
 * 
 * @author Lennaert van den Brink <l.o.vandenbrink@student.tudelft.nl>
 * @author Hans Janmaat <c.j.janmaat@student.tudelft.nl>
 * @author Cees-Willem Hofstede <c.w.p.l.j.hofstede@student.tudelft.nl>
 * @author Berend Klein Haneveld <b.a.kleinhaneveld@student.tudelft.nl>
 *
 */
public class OpponentModel {
	private Group4 agent;
	private ArrayList<Issue> issues;
	private ArrayList<ArrayList<Double>> issuesTracker;
	
	public ArrayList<ExtendedBid> bids; 		// All bids done by the opponent
	public ArrayList<ExtendedBid> uniqueBids;	// All unique bids done by the opponent
	
	public OpponentModel(Group4 agent) {
		this.agent = agent;
		issues = agent.utilitySpace.getDomain().getIssues();
		bids = new ArrayList<ExtendedBid>();
		uniqueBids = new ArrayList<ExtendedBid>();

		
		initializeTracker();
	}
	
	/**
	 * This function is called when a next session starts with the same opponent
	 */
	public void init() {
		initializeTracker();
	}
	
	/**
	 * 'Stores' opponent action in the matrix and applies normalization
	 * 
	 * @param action Action from opponent
	 */
	public void opponentOffered(Offer offer) {
		Bid opponentBid = offer.getBid();
		
		try {
			addBidToIssuesTracker(opponentBid, 1.0);
		} catch (Exception e) {
			agent.error(e.getStackTrace().toString());
		}
	}
	
	/**
	 * Returns the bid that has a higher utility than lowestAllowedUtility that has the 
	 * highest possible value for our opponent according to our issue tracker. This bid
	 * should be quite near the Pareto line
	 * @param lowestAllowedUtility Minimum utility value that the bid should have
	 * @return Bid that is near the Pareto line with utility higher than lowestAllowedUtility
	 */
	public Bid getBestNextBid(double lowestAllowedUtility) {
		// For all bids in rankedbids with utility higher than lowestAllowedUtility:
		// decide which bid will maximize the utility of the opponent
		
		// In order to maximize the utility, the sum of the values from the matrix can be used
		// So the next for loop searches for the bid that has the biggest value for our opponent
		// while not exceeding the specified utility: lowestAllowedUtility
		int maxIndex = getIndexOfFirstBidWithLowerUtilityThanTheProvidedUtility(lowestAllowedUtility);
		Bid result = agent.rankedBids.get(0).bid;
		double resultValue = 0.0;
		for (int i = 0; i < maxIndex; i++) {
			Bid bid = agent.rankedBids.get(i).bid;
			double valueOfBid = valueOfBidInIssueTracker(bid);
			if (valueOfBid > resultValue) {
				resultValue = valueOfBid;
				result = bid;
			}
		}
		
		return result;
	}
	
	/**
	 * Returns the value of a bid according to the issue tracker. Value is calculated
	 * as the sum of the values that the bid represents in the issue tracker
	 * @param bid Bid of which the value needs to be found
	 * @return a number that represents the 'value' of the bid
	 */
	private double valueOfBidInIssueTracker(Bid bid) {
		BidIssue bidIssue = new BidIssue(bid, issues);
		
		double result = 0.0;
		
		for (int i = 0; i < issues.size(); i++) {
			ArrayList<Double> issue = issuesTracker.get(i);
			result += issue.get(bidIssue.issueIndices[i]);
		}
		
		return result;
	}
	
	/**
	 * Create string representation from issue tracker
	 * @return Description of issue tracker matrix
	 */
	public String toString() {
		String result = "";
		
		int numberOfIssues = issues.size();
		
		for (int i = 0; i < numberOfIssues; i++) {
			result += "[";
			ArrayList<Double> issueValues = issuesTracker.get(i);
			for (int j = 0; j < issueValues.size(); j++) {
				result += issueValues.get(j) + (j == issueValues.size()-1 ? "" : ", ");
			}
			result += "]\n";
		}
		
		return result;
	}
	
	/**
	 * Checks the opponent's bids to see if the given bid is unique
	 * @param bid to be checked
	 * @return true when bid is not in opponent's bids
	 */
	public boolean isUniqueBid(ExtendedBid bid) {
		boolean containsBid = false;
		for (int j = 0; j < uniqueBids.size() && !containsBid; j++) {
			ExtendedBid otherBid = uniqueBids.get(j);
			containsBid = otherBid.equals(bid);
		}
		
		return !containsBid;
	}
	
	
	/**
	 * Does what it says basically
	 * @param utility upper limit?
	 * @return Index of first bid with lower utility than the provided utility
	 */
	public int getIndexOfFirstBidWithLowerUtilityThanTheProvidedUtility(double utility) {
		int result = -1;
		
		for (int i = 0; i < agent.rankedBids.size() && result < 0; i++) {
			if (agent.rankedBids.get(i).utility < utility)
				result = i;
		}
		
		// result must be at least 0
		return result >= 0 ? result : 0;
	}
	
	/**
	 * @param index of bid from where to search from
	 * @param epsilon range of how big the difference between bids may be
	 * @return Index of bid that deviates the most from the bid at the provided index
	 * within range of that utility + epsilon
	 * If no bid is found within that range, the provided index is returned
	 */
	public int getIndexOfBidWithinRangeOfProvidedIndex(int index, double epsilon) {
		int result = index; 
		
		for (int i = index; epsilon > 0.0 ? i >= 0 : i < agent.rankedBids.size();) {
			if (epsilon > 0) {
				if (agent.rankedBids.get(i).utility > agent.rankedBids.get(index).utility + epsilon) {
					result = i+1;
					break;
				}
			} else {
				if (agent.rankedBids.get(i).utility < agent.rankedBids.get(index).utility + epsilon) {
					result = i-1;
					break;
				}
			}
			i = epsilon > 0.0 ? i-1 : i+1;
		}
		
		return result >= 0 ? result : index;
	}
	
	/* Private stuff */
	
	/**
	 * Initializes/resets the bid tracker matrix
	 */
	private void initializeTracker() {
		issuesTracker = new ArrayList<ArrayList<Double>>();
		for (int i = 0; i < issues.size(); i++) {
			ArrayList<Double> issueValues = new ArrayList<Double>();
			IssueDiscrete issue = (IssueDiscrete)issues.get(i);
			for (int j = 0; j < issue.getValues().size(); j++) {
				// Initialize with value 1.0 so that each combination/bid will have a value
				issueValues.add(1.0);
			}
			issuesTracker.add(issueValues);
		}
	}
	
	/**
	 * Adds the given factor to the bid's values in the issues tracker
	 * @param bid from the opponent
	 * @param factor that is added to the values of the bid in the tracker
	 * @throws Exception because of Bid.getValue(...)
	 */
	private void addBidToIssuesTracker(Bid bid, Double factor) throws Exception {
		// Count the number of issues
		int numberOfIssues = issues.size();
		
		for (int i = 0; i < numberOfIssues; i++) {
			// Get the current double values (row) for issue i from the 'matrix' with 'probability' values
			ArrayList<Double> issueValues = issuesTracker.get(i);
			
			// TODO: Warning: start of workaround for bug where issue index is not issue index that should
			// be provided to getValue(int index) of Bid!
			int firstPossibleIndexOfValueInBid = -1;
			boolean succeeded = false;
			while (!succeeded) {
				firstPossibleIndexOfValueInBid++;
				try {
					bid.getValue(firstPossibleIndexOfValueInBid);
					succeeded = true;
				} catch (Exception e) {
					// Nothing to see here, Bid.getValue() not correctly implemented
				}
			}
			
			// Get the ValueDiscrete of the bid
			ValueDiscrete valueOfBid = (ValueDiscrete)bid.getValue(i+firstPossibleIndexOfValueInBid);
			// Figure out the index in the row of double values to add to
			int indexOfValue = indexOfValueInIssue(valueOfBid, (IssueDiscrete)issues.get(i));
			
			if (indexOfValue < 0)
				agent.info("Error: bids en values zijn niet goed geprogrammeerd");
			
			// Add a certain value to this value 
			issueValues.set(indexOfValue, issueValues.get(indexOfValue) + factor);
		}
	}
	
	/**
	 * Find the index in the row of a certain issue from the domain
	 * @param value from a bid
	 * @param issue that should contain this value 
	 * @return index
	 */
	private int indexOfValueInIssue(ValueDiscrete value, IssueDiscrete issue) {
		int indexInIssue = -1;
		
		ArrayList<ValueDiscrete> values = issue.getValues();
		for (int i = 0; i < issue.getNumberOfValues() && indexInIssue == -1; i++) {
			if (values.get(i).value.equals(value.value)) {
				indexInIssue = i;
			}
		}
		
		// Print an error if no index can be found
		if (indexInIssue < 0)
			agent.error("Error: " + value + " not found in: " + issues.toString());
		
		return indexInIssue;
	}
}
