package negotiation.group4;

import negotiator.Bid;
import negotiator.actions.Accept;
import negotiator.actions.Action;
import negotiator.actions.EndNegotiation;
import negotiator.actions.Offer;

/**
 * Strategy is used to return an action to the agent.
 * 
 * @author Lennaert van den Brink <l.o.vandenbrink@student.tudelft.nl>
 * @author Hans Janmaat <c.j.janmaat@student.tudelft.nl>
 * @author Cees-Willem Hofstede <c.w.p.l.j.hofstede@student.tudelft.nl>
 * @author Berend Klein Haneveld <b.a.kleinhaneveld@student.tudelft.nl>
 *
 */
public class Strategy {
	// Keep a reference to the agent
	Group4 agent;
	
	double opponentsFirstUtility 	= -1.0;
	double highestUtility 			= -1.0;
	double factor					= 0.95;
	double timeOfFirstBid			= 0.0;		// measured in milliseconds
	double waitingFraction			= 0.90;
	double acceptFactor				= 0.60;
	
	// create bids with a certain offset if this number exceeds 20, reset after 5 random bids or something
	int maxIdenticalOpponentBids = 10; // TODO: shouldn't this be based on total number of possible bids?
	int identicalOpponentBids = 0;
	int numberOfRandomBids = 0;
	int randomIndexMaximumOffset = 5;
	Bid lastCachedOpponentBid = null;
	
	public Strategy(Group4 agent) {
		this.agent = agent;
	}
	
	/**
	 * Returns the fraction of the time the agent should wait
	 * So don't forget to multiply with the remaining time
	 * 
	 * @param bidNumber 
	 * @return
	 */
	double howLongShouldWeWait(int bidNumber) {
		double result = 0.0;
		
		double discountFactor = agent.utilitySpace.getDiscountFactor();
		double discountWaitFactor = 1.0; // Default value for the discount factor
		// If there is a discount factor, calculate a factor dependent on the amount of discount
		// Changed from 1/3 to 1/2 because on discounted domains opponents get scared quite easily, so
		// we don't have to wait that long most probably
		if (discountFactor > 0.0)
			discountWaitFactor = Math.max(Math.pow(discountFactor, 1.0/2.0), 0.1);
		
		double reservationValue = agent.utilitySpace.getReservationValueUndiscounted();
		double discountReservationFactor = 1.0;
		if (reservationValue > 0.0)
			discountReservationFactor = 0.9 + 0.1 * reservationValue;
		
		result = (discountReservationFactor * discountWaitFactor * waitingFraction) / Math.pow(bidNumber, 2.3);
		
		return result;
	}
	
	public Action getAction() throws Exception {
		Action action = null;
		
		// Keep track of the variaty of the opponent bids
		Bid opponentBid = null;
		if (agent.opponentModel.bids.size() > 0) // Set the opponent bid if there are any
			opponentBid = agent.opponentModel.bids.get(agent.opponentModel.bids.size()-1).bid;
		// Save the last bid if it not the same as last time, otherwise increase the number of identical bids
		if (lastCachedOpponentBid == null || !opponentBid.equals(lastCachedOpponentBid)) {
			identicalOpponentBids = 0;
			lastCachedOpponentBid = opponentBid;
		} else
			identicalOpponentBids++;
		
		// Get important time values
		long totalMilliSeconds = agent.timeline.getTotalMiliseconds();
		double elapsedMilliSeconds = agent.timeline.getElapsedMilliSeconds();
		
		// For the first time: wait until 'a certain percentage' of the available time has passed to do a bid
		if (agent.thisAgentBids.size() == 0) {
			return new Offer(agent, agent.getOfferBid(0).bid);
		}
		
		// Calculate how far we are in the session
		double timeRatio = (elapsedMilliSeconds - timeOfFirstBid) / (totalMilliSeconds - timeOfFirstBid);
		double combinedFactor = factor * (1 - Math.pow(timeRatio, 2));
		
		// Save the highest possible utility value
		if (highestUtility < 0)
			highestUtility = agent.rankedBids.get(0).utility;
		// On first bid of other agent, save the utility value of that bid
		if (opponentsFirstUtility < 0 && agent.opponentModel.bids.size() > 0)
			opponentsFirstUtility = agent.opponentModel.bids.get(0).utility;
		// If the opponent has not made an offer, make the bid with the highest utility
		if (opponentsFirstUtility < 0)
			return new Offer(agent, agent.getOfferBid(0).bid);
		
		/* 
		 * Keep/update a factor/ratio that tells how much can be proceeded to the opponent
		 * Example: own value is 1, opponent's value is 0.4, ratio is 0.5, than all bids
		 * can be done that are higher than 0.4 + 0.5 * (1 - 0.4)
		 * This factor gets lower over time so that it 'concedes' a bit to the opponent
		 */
		double allTimeLow = opponentsFirstUtility + acceptFactor * (highestUtility - opponentsFirstUtility);
		double minimumUtility = allTimeLow + (combinedFactor * (highestUtility - allTimeLow));
		
		// If the opponent's bid has an acceptable utility, accept
		double lastOpponentsUtility = agent.opponentModel.bids.get(agent.opponentModel.bids.size()-1).utility; 
		if (lastOpponentsUtility > minimumUtility) {
			agent.info("Factor: " + combinedFactor);
			agent.info("Lowest acceptable utility: " + minimumUtility);
			agent.info("Time: " + agent.timeline.getElapsedSeconds());
			return new Accept(agent);
		}
		
		// TODO: Do some check, fails against some agents!
		// Check to see if the opponent reacts to our bids, if not, try some random bids with about the same utility as the minimum utility
		if (identicalOpponentBids > maxIdenticalOpponentBids) {
			// Find a min and max index of bids that are within 'epsilon' of our minimum bid.
			double epsilon = 0.01; // TODO: this is quite low for domains with a small amount of bids
			
			int minIndex = agent.opponentModel.getIndexOfBidWithinRangeOfProvidedIndex(agent.lastBidIndex, epsilon);
			int maxIndex = agent.opponentModel.getIndexOfBidWithinRangeOfProvidedIndex(agent.lastBidIndex, -epsilon);
			
			// Then choose a random bid within that range
			int randomIndex = minIndex + (int)(Math.random() * (maxIndex - minIndex));
			/*
			agent.info("---------------------");
			agent.info("Max index: " + maxIndex 			+ " at " + agent.rankedBids.get(maxIndex).utility);
			agent.info("Ori index: " + agent.lastBidIndex	+ " at " + agent.rankedBids.get(agent.lastBidIndex).utility);
			agent.info("Min index: " + minIndex				+ " at " + agent.rankedBids.get(minIndex).utility);
			agent.info("---------------------");
			*/
			if(randomIndex < 0) // just a failsafe
				randomIndex = 0;
			numberOfRandomBids++;
			agent.info("randomIndex: " + randomIndex);
			if (randomIndex < agent.rankedBids.size()) {
				ExtendedBid bid = agent.getOfferBid(randomIndex);
				// Only create a new action if the bid has not been done by the opponent
				if (!opponentBidsContainBid(bid) && bid.utility > getHighestOpponentUtility())
					action = new Offer(agent, bid.bid);
				
				// Reset the counter so that you only do 5 random bids or so
				if (numberOfRandomBids > 10)
					identicalOpponentBids = 0;
			}
		} else {
			numberOfRandomBids = 0;
		}

		// If action has not been set, we need to find out what to do next
		if (action == null) {
			// Get the best possible bid according to the opponent model within our current minimum utility
			Bid bestBid = agent.opponentModel.getBestNextBid(minimumUtility);
			assert(agent.utilitySpace.getUtility(bestBid) >= minimumUtility);
			double utilWithDiscount = agent.utilitySpace.getUtilityWithDiscount(bestBid, agent.timeline.getTime());
			double resvWithDiscount = agent.utilitySpace.getReservationValueWithDiscount(agent.timeline.getTime());
			if (utilWithDiscount <= resvWithDiscount)
				return new EndNegotiation(agent.getAgentID());
			else
				action = new Offer(agent, bestBid);
		}
		
		// Wait before doing another action
		double waitTimeFraction = howLongShouldWeWait(agent.thisAgentBids.size());
		double sleepTimeFraction = waitTimeFraction * (1.0 - elapsedMilliSeconds / totalMilliSeconds);
		
		// What will be the reservation value at time of next bid?
		double timeAtNextBid = (elapsedMilliSeconds / totalMilliSeconds) + sleepTimeFraction;
		if (action instanceof Offer) {
			Offer offer = (Offer)action;
			double value = agent.utilitySpace.getUtilityWithDiscount(offer.getBid(), timeAtNextBid);
			// If the discounted value of our next bid is lower than the reservation value now, just end the negotiation
			if (value < agent.utilitySpace.getReservationValueWithDiscount(agent.timeline.getTime()))
				return new EndNegotiation(agent.getAgentID());
		}
		
		// Sleep, my child...
		if (sleepTimeFraction > 0.0005) {
			agent.info("Agent will sleep for " + sleepTimeFraction * 180 + " seconds...");
			double sleptTime = 0.0;
			while (sleepTimeFraction > 0.0) {
				double sleepF = 1.0 / 18.0 < sleepTimeFraction ? 1.0 / 18.0 : sleepTimeFraction;
				agent.sleep(sleepF);
				sleepTimeFraction -= sleepF;
				sleptTime += sleepF;
				agent.info("Slept for " + Math.round(sleptTime * 180) + " seconds...");
			}
		}
		
		// Save the time of our first bid, so that it can be used to properly calculate time ratio later on
		if (agent.thisAgentBids.size() == 1)
			timeOfFirstBid = agent.timeline.getElapsedMilliSeconds();
		
		return action;
	}
	
	protected double getHighestOpponentUtility() {
		double result = -1.0;
		
		for (int i = 0; i < agent.opponentModel.bids.size(); i++)
			result += agent.opponentModel.bids.get(i).utility > result ? agent.opponentModel.bids.get(i).utility : 0.0;
		
		return result;
	}
	
	/**
	 * Returns whether the bids from the opponent contain
	 * a bid that is the same as the given bid
	 * @param extendedBid Bid that may be in opponent bids
	 * @return boolean whether the bid is contained in the opponent's bids
	 */
	protected boolean opponentBidsContainBid(ExtendedBid extendedBid) {
		boolean result = false;
		
		for (int i = 0; i < agent.opponentModel.bids.size() && !result; i++)
			result = agent.opponentModel.bids.get(i).equals(extendedBid);
		
		return result;
	}

	public void reset() {
		opponentsFirstUtility = -1.0;
		highestUtility = -1.0;
		factor = 0.95;
		timeOfFirstBid = 0.0;
	}
	
}
